#include <cassert>

#include <cstddef>
using std::size_t;

#include <cstdlib>
using std::exit;

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <vector>
using std::vector;

#include <cmath>
using std::sin;
using std::fabs;

#include <iostream>
using std::cout;
using std::endl;
using std::ostream;

#include <iomanip>
using std::setw;
using std::setfill;

#include <fstream>
using std::ofstream;

#include <sstream>
using std::stringstream;

const double L = 10.0;
const double halfL = L/2;
const double halfL2 = halfL*halfL;

const unsigned int N = 1000;
const unsigned int N1 = N+1;
vector<double> u2(N1);
vector<double> u1(N1);
vector<double> u0(N1);

vector<double> u(N1);
vector<double> v1(N1);
vector<double> v0(N1);
vector<double> theta1(N1);
vector<double> theta0(N1);

// x = n*dx, n=0,1,..,N;
// u(0) = u(L) = 0  =>  u[0] = u[N] = 0.0;

const double dx = L/N;
const double c = 0.5;
const double dt = 0.0075;
						
const bool use_wave = true;
const bool use_advection = false;
const bool implicit = false;
const bool slow_motion = false;

inline double mode(unsigned int n, double x) { return sin(n*M_PI*x/L); }

double lowest_harmonic(double x)
{
	return mode(1,x);
}

double harmonics(double x)
{
	return mode(1,x)+0.1*mode(2,x) - 0.05*mode(3,x) + 0.01*mode(4,x);
}

double square(double x)
{
	return (x > 2.0 and x < 3.0 ? 1.0 : 0.0);
}

double pluck(double x)
{
	const double y = x-halfL;
	return exp(-y*y)-exp(-halfL2);
}

enum direction { LEFT, RIGHT, STATIONARY };
void initialize(double (&f)(double), direction d)
{
	double shift;
	if (d == LEFT)
		shift = c*dt;
	else if (d == RIGHT)
		shift = -c*dt;
	else if (d == STATIONARY)
		shift = 0.0;
	else
		assert(false);
	u2[0] = u2[N] = 0.0;
	u2[1] = u2[N-1] = 0.0;
	for (size_t i = 1; i < N; ++i)
	{
		u1[i] = f(i*dx);
		u0[i] = f(i*dx-shift);
	}
	u1[0] = u1[N] = 0.0;
	u0[0] = u0[N] = 0.0;

	u = u1;
	for (size_t i = 0; i < N1; ++i)
		v0[i] = v1[i] = (u1[i]-u0[i])/dt;
	theta0[0] = -c*(u0[1]-u0[0])/dx;
	theta1[0] = -c*(u1[1]-u1[0])/dx;
	theta0[N] = -c*(u0[N]-u0[N-1])/dx;
	theta1[N] = -c*(u1[N]-u1[N-1])/dx;
	v1[0] = v1[N] = 0.0;
	v0[0] = v0[N] = 0.0;
	for (size_t i = 1; i < N; ++i)
	{
		theta1[i] = -c*(u1[i+1]-u1[i-1])/(2.0*dx);
		theta0[i] = -c*(u0[i+1]-u0[i-1])/(2.0*dx);
	}
}

const double s = c*dt/dx;
const double s2 = s*s;
void update(void)
{
	if (use_wave)
	{
		// u(n+1,i)-u(n,i) = 
		for (size_t i = 1; i < N; ++i)
		
			u2[i] = 2.0*(1.0-s2)*u1[i] - u0[i] + s2*(u1[i+1] + u1[i-1]);
		// u2 u(n)     u(n)   * u(n-2)           xxx
		// u1 u(n-1) * u(n-2) * u(n)    n->n+1   u(n-1)
		// u0 u(n-2) * u(n-1)   u(n-1)           u(n-2)
		swap(u0,u1);
		swap(u1,u2);
	}
	
	//explict 4-point
	if (use_advection and !implicit)
	{
		v1[0] = v0[0] + s*(theta0[1]-theta0[0]);
		for (size_t i = 1; i < N; ++i)
			v1[i] = v0[i] - 0.5*s*(theta0[i+1]-theta0[i-1]);
		v1[N] = v0[N] + s*(theta0[N]-theta0[N-1]);
		theta1[0] = theta0[0] - s*(v0[1]-v0[0]);
		for (size_t i = 1; i < N; ++i)
			theta1[i] = theta0[i] - 0.5*s*(v0[i+1]-v0[i-1]);
		theta1[N] = theta0[N] - s*(v0[N]-v0[N-1]);
		for (size_t i = 0; i < N1; ++i)
			u[i] += dt*0.5*(v0[i]+v1[i]);
		swap(v0,v1);
		swap(theta0,theta1);
	}

	// implicit 6-point (Crank-Nicholson)
	if (use_advection and implicit)
	{
		// Jacobi method for self-consistency loop
		for (size_t count = 0; count < 100; ++count)
		{
			// +!
			v1[0] = v0[0] + 0.5*s*(theta0[1]-theta0[0]+theta1[1]-theta1[0]);
			for (size_t i = 1; i < N; ++i)
			//for (size_t i = N-1; i >= 1; --i)
				v1[i] = v0[i] - 0.25*s*(theta0[i+1]-theta0[i-1]+theta1[i+1]-theta1[i-1]);
			// +!
			v1[N] = v0[N] + 0.5*s*(theta0[N]-theta0[N-1]+theta1[N]-theta1[N-1]);
			theta1[0] = theta0[0] - 0.5*s*(v0[1]-v0[0]+v1[1]-v1[0]);
			for (size_t i = 1; i < N; ++i)
			//for (size_t i = N-1; i >= 1; --i)
				theta1[i] = theta0[i] - 0.25*s*(v0[i+1]-v0[i-1]+v1[i+1]-v1[i-1]);
			theta1[N] = theta0[N] - 0.5*s*(v0[N]-v0[N-1]+v1[N]-v1[N-1]);
		}
		for (size_t i = 0; i < N1; ++i)
			u[i] += dt*0.5*(v0[i]+v1[i]);

		for (size_t i = 0; i < N1; ++i)
		{
			v0[i] = v1[i];
			theta0[i] = theta1[i];
		}
	}
}

void display()
{
	glClear(GL_COLOR_BUFFER_BIT);
	glBegin(GL_LINES);
		glColor3f(1.0, 0.2, 0.3); 
		glVertex2f(-0.5,0.25);
		glVertex2f(0.5,0.25);
	glEnd();
	glBegin(GL_LINES);
		glColor3f(1.0, 0.2, 0.3); 
		glVertex2f(-0.5,-0.25);
		glVertex2f(0.5,-0.25);
	glEnd();
	if (use_wave)
	{
		glBegin(GL_LINE_STRIP);
			glColor3f(1.0, 1.0, 1.0); 
			for (size_t i = 0; i < N1; ++i)
			{
				const double x = double(i)/N1-0.5;
				const double y = 0.25*u2[i];
				if (fabs(y) < 1.0)
					glVertex2f(x,y);
				else
				{
					glEnd();
					glBegin(GL_LINE_STRIP);
						glColor3f(1.0, 1.0, 1.0); 
				}
			}
		glEnd();
	}
	if (use_advection)
	{
		glBegin(GL_LINE_STRIP);
			glColor3f(0.0, 1.0, 1.0); 
			for (size_t i = 0; i < N1; ++i)
			{
				const double x = double(i)/N1-0.5;
				const double y = 0.25*u[i];
				if (fabs(y) < 1.0)
					glVertex2f(x,y);
				else
				{
					glEnd();
					glBegin(GL_LINE_STRIP);
						glColor3f(0.0, 1.0, 1.0); 
				}
			}
		glEnd();
	}
	glBegin(GL_POLYGON);
		glColor3f(0.0, 0.0, 1.0); 
		glVertex2f(-0.5,0.025);
		glVertex2f(-0.5,-0.025);
		glVertex2f(-0.575,-0.025);
		glVertex2f(-0.575,0.025);
	glEnd();
	glBegin(GL_POLYGON);
		glColor3f(0.0, 0.0, 1.0); 
		glVertex2f(0.5,0.025);
		glVertex2f(0.5,-0.025);
		glVertex2f(0.575,-0.025);
		glVertex2f(0.575,0.025);
	glEnd();
	glFlush();
	glutSwapBuffers();
}

const int delay = (slow_motion ? 50 : (use_advection and implicit ? 1 : 30)); // milliseconds
void animate(int)
{
	if (slow_motion)
		update();
	else for (size_t i = 0; i < 20; ++i)
		update();
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
}

void keyboard(unsigned char key, int x, int y)
{
	if (key == 'q')
		exit(0);
}

int main(int argc, char **argv)
{
	// initial values

	//initialize(lowest_harmonic,STATIONARY);
	//initialize(harmonics,STATIONARY);

	initialize(pluck,RIGHT);
	//initialize(pluck,LEFT);
	//initialize(pluck,STATIONARY);

	//initialize(square,RIGHT);
	//initialize(square,LEFT);
	//initialize(square,STATIONARY);

	cout << "Simulating with s = " << s << endl;
	
	glutInit(&argc,argv);
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (640, 640);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("Vibrating String");
	glutDisplayFunc(display);
   glutKeyboardFunc(keyboard);
	glutTimerFunc(delay,animate,0);
	glutMainLoop();
}
